@@ -36,6 +36,8 @@
 #include <QtVirtualKeyboard/qvirtualkeyboardinputengine.h>
 
 #include <QGuiApplication>
+#include <QtQuick/qquickitem.h>
+#include <QtQuick/qquickwindow.h>
 #include <QtGui/qpa/qplatformintegration.h>
 #include <QtGui/private/qguiapplication_p.h>
 
@@ -51,6 +53,8 @@ bool operator==(const QInputMethodEvent::Attribute &attribute1, const QInputMeth
 
 using namespace QtVirtualKeyboard;
 
+const bool QtVirtualKeyboard::QT_VIRTUALKEYBOARD_FORCE_EVENTS_WITHOUT_FOCUS = qEnvironmentVariableIsSet("QT_VIRTUALKEYBOARD_FORCE_EVENTS_WITHOUT_FOCUS");
+
 QVirtualKeyboardInputContextPrivate::QVirtualKeyboardInputContextPrivate(QVirtualKeyboardInputContext *q_ptr) :
     QObject(nullptr),
     q_ptr(q_ptr),
@@ -211,6 +215,13 @@ bool QVirtualKeyboardInputContextPrivate::hasEnterKeyAction(QObject *item) const
     return item != nullptr && qmlAttachedPropertiesObject<EnterKeyAction>(item, false);
 }
 
+void QVirtualKeyboardInputContextPrivate::registerInputPanel(QObject *inputPanel)
+{
+    VIRTUALKEYBOARD_DEBUG() << "QVirtualKeyboardInputContextPrivate::registerInputPanel():" << inputPanel;
+    Q_ASSERT(!this->inputPanel);
+    this->inputPanel = inputPanel;
+}
+
 void QVirtualKeyboardInputContextPrivate::hideInputPanel()
 {
     platformInputContext->hideInputPanel();
@@ -257,10 +268,41 @@ void QVirtualKeyboardInputContextPrivate::forceCursorPosition(int anchorPosition
 
 void QVirtualKeyboardInputContextPrivate::onInputItemChanged()
 {
-    if (!inputItem() && !activeKeys.isEmpty()) {
-        // After losing keyboard focus it is impossible to track pressed keys
-        activeKeys.clear();
-        clearState(State::KeyEvent);
+    if (QObject *item = inputItem()) {
+        if (QQuickItem *vkbPanel = qobject_cast<QQuickItem*>(inputPanel)) {
+            if (QQuickItem *quickItem = qobject_cast<QQuickItem*>(item)) {
+                const QVariant isDesktopPanel = vkbPanel->property("desktopPanel");
+                /*
+                    For integrated keyboards, make sure it's a sibling to the overlay. The
+                    high z-order will make sure it gets events also during a modal session.
+                */
+                if (isDesktopPanel.isValid() && !isDesktopPanel.toBool()) {
+                    if (QQuickWindow *quickWindow = quickItem->window()) {
+                        QQuickItem *overlay = quickWindow->property("_q_QQuickOverlay").value<QQuickItem*>();
+                        if (overlay && overlay->isVisible()) {
+                            if (vkbPanel->parentItem() != overlay->parentItem()) {
+                                inputPanelParentItem = vkbPanel->parentItem();
+                                inputPanelZ = vkbPanel->z();
+                                vkbPanel->setParentItem(overlay->parentItem());
+                                vkbPanel->setZ(overlay->z() + 1);
+                            }
+                        } else {
+                            if (QQuickItem *oldParentItem = qobject_cast<QQuickItem *>(inputPanelParentItem.data())) {
+                                vkbPanel->setParentItem(oldParentItem);
+                                vkbPanel->setZ(inputPanelZ);
+                                inputPanelParentItem = nullptr;
+                            }
+                        }
+                    }
+                }
+            }
+        }
+    } else {
+        if (!activeKeys.isEmpty()) {
+            // After losing keyboard focus it is impossible to track pressed keys
+            activeKeys.clear();
+            clearState(State::KeyEvent);
+        }
     }
     clearState(State::InputMethodClick);
 }
@@ -364,7 +406,7 @@ void QVirtualKeyboardInputContextPrivate::update(Qt::InputMethodQueries queries)
     bool newCursorPosition = cursorPosition != this->cursorPosition;
     bool newAnchorRectangle = anchorRectangle != this->anchorRectangle;
     bool newCursorRectangle = cursorRectangle != this->cursorRectangle;
-    bool selectionControlVisible = platformInputContext->isInputPanelVisible() && (cursorPosition != anchorPosition) && !inputMethodHints.testFlag(Qt::ImhNoTextHandles);
+    bool selectionControlVisible = platformInputContext->evaluateInputPanelVisible() && (cursorPosition != anchorPosition) && !inputMethodHints.testFlag(Qt::ImhNoTextHandles);
     bool newSelectionControlVisible = selectionControlVisible != this->selectionControlVisible;
 
     QRectF inputItemClipRect = imQueryEvent.value(Qt::ImInputItemClipRectangle).toRectF();
@@ -478,6 +520,7 @@ bool QVirtualKeyboardInputContextPrivate::filterEvent(const QEvent *event)
     QEvent::Type type = event->type();
     if (type == QEvent::KeyPress || type == QEvent::KeyRelease) {
         const QKeyEvent *keyEvent = static_cast<const QKeyEvent *>(event);
+        const int key = keyEvent->key();
 
         // Keep track of pressed keys update key event state
         if (type == QEvent::KeyPress)
@@ -491,7 +534,6 @@ bool QVirtualKeyboardInputContextPrivate::filterEvent(const QEvent *event)
             setState(State::KeyEvent);
 
 #ifdef QT_VIRTUALKEYBOARD_ARROW_KEY_NAVIGATION
-        int key = keyEvent->key();
         if ((key >= Qt::Key_Left && key <= Qt::Key_Down) || key == Qt::Key_Return) {
             if (type == QEvent::KeyPress && platformInputContext->isInputPanelVisible()) {
                 activeNavigationKeys += key;
@@ -506,9 +548,26 @@ bool QVirtualKeyboardInputContextPrivate::filterEvent(const QEvent *event)
 #endif
 
         // Break composing text since the virtual keyboard does not support hard keyboard events
-        if (!preeditText.isEmpty())
-            commit();
+        if (!preeditText.isEmpty()) {
+            if (type == QEvent::KeyPress && (key == Qt::Key_Delete || key == Qt::Key_Backspace)) {
+                reset();
+                Q_Q(QVirtualKeyboardInputContext);
+                q->clear();
+                return true;
+            } else {
+                commit();
+            }
+        }
     }
+#ifdef QT_VIRTUALKEYBOARD_ARROW_KEY_NAVIGATION
+    else if (type == QEvent::ShortcutOverride) {
+        const QKeyEvent *keyEvent = static_cast<const QKeyEvent *>(event);
+        int key = keyEvent->key();
+        if ((key >= Qt::Key_Left && key <= Qt::Key_Down) || key == Qt::Key_Return)
+            return true;
+    }
+#endif
+
     return false;
 }
 
